import os
import functools


class DynamicDramatiq:

    def __init__(self) -> None:
        self.broker = None
        self.func_to_actor_kwargs = {}
        self.func_to_actor = {}

    def actor(self, **actor_kwargs):

        def decorator(func):
            self.func_to_actor_kwargs[func] = actor_kwargs

            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                actor = self.func_to_actor.get(func)
                if actor is None:
                    raise RuntimeError('Broker not set.')

                return actor.send(*args, **kwargs)

            return wrapped

        return decorator

    def set_broker(self, broker):
        import dramatiq  # pylint: disable=import-outside-toplevel

        self.broker = broker
        dramatiq.set_broker(broker)

        self.func_to_actor = {}
        for func, actor_kwargs in self.func_to_actor_kwargs.items():
            self.func_to_actor[func] = dramatiq.actor(**actor_kwargs)(func)


dynamic_dramatiq = DynamicDramatiq()  # pylint: disable=invalid-name

# If set, enter worker mode.
_REDIS_BROKER_PORT = os.getenv('DYNAMIC_DRAMATIQ_REDIS_BROKER_PORT')
if _REDIS_BROKER_PORT:
    # 1. Load actors.
    import private_pypi.workflow  # pylint: disable=unused-import
    from private_pypi.backend import BackendInstanceManager
    BackendInstanceManager()

    # 2. Connect to broker.
    from dramatiq.brokers.redis import RedisBroker
    REDIS_BROKER = RedisBroker(host='localhost', port=_REDIS_BROKER_PORT)
    dynamic_dramatiq.set_broker(REDIS_BROKER)



import base64
from dataclasses import dataclass
from datetime import datetime
import json
import hashlib
import os
import os.path
import re
import shutil
from typing import Callable, TextIO, Any
import uuid
import zlib

from cryptography.fernet import Fernet
from filelock import FileLock
import toml


def write_toml(path, struct):
    with open(path, 'w') as fout:
        fout.write(toml.dumps(struct))


def read_toml(path):
    with open(path) as fin:
        return toml.loads(fin.read())


def locked_read_file(lock_path, file_path, timeout=-1):
    try:
        with FileLock(lock_path, timeout=timeout):
            if not os.path.exists(file_path):
                return True, None
            with open(file_path) as fin:
                return True, fin.read()
    except TimeoutError:
        return False, ''


def locked_read_toml(lock_path, file_path, timeout=-1):
    status, text = locked_read_file(lock_path, file_path, timeout=timeout)
    struct = None
    if status and text is not None:
        struct = toml.loads(text)
    return status, struct


def locked_write_file(lock_path, file_path, text, timeout=-1):
    try:
        with FileLock(lock_path, timeout=timeout):
            with open(file_path, 'w') as fout:
                fout.write(text)
            return True
    except TimeoutError:
        return False


def locked_write_toml(lock_path, file_path, struct, timeout=-1):
    return locked_write_file(lock_path, file_path, toml.dumps(struct), timeout=timeout)


def locked_copy_file(lock_path, src_path, dst_path, timeout=-1):
    try:
        with FileLock(lock_path, timeout=timeout):
            shutil.copyfile(src_path, dst_path)
            return True
    except TimeoutError:
        return False


def file_lock_is_busy(lock_path):
    flock = FileLock(lock_path)
    busy = False
    try:
        flock.acquire(timeout=0.1, poll_intervall=0.05)
    except TimeoutError:
        busy = True
    finally:
        flock.release()
    return busy


@dataclass
class LockedFileLikeObject(TextIO):
    # pylint: disable=abstract-method
    lock_path: str
    write_func: Callable

    def write(self, s: str) -> int:
        with FileLock(self.lock_path):
            self.write_func(s)
        return 0


def normalize_distribution_name(name: str) -> str:
    # https://www.python.org/dev/peps/pep-0503/#normalized-names
    return re.sub(r'[-_.]+', '-', name).lower()


def update_hash_algo_with_file(path: str, hash_alog: Any) -> None:
    with open(path, 'rb') as fin:
        # 64KB block.
        for block in iter(lambda: fin.read(65536), b''):
            hash_alog.update(block)


def git_hash_sha(path: str) -> str:
    # https://stackoverflow.com/questions/5290444/why-does-git-hash-object-return-a-different-hash-than-openssl-sha1
    sha1_algo = hashlib.sha1()
    size = os.path.getsize(path)
    sha1_algo.update(f'blob {size}\0'.encode())
    update_hash_algo_with_file(path, sha1_algo)
    return sha1_algo.hexdigest()


def get_secret_key():
    secret_key = os.getenv('PRIVATE_PYPI_SECRET_KEY')
    if secret_key is None:
        secret_key = str(uuid.getnode())
    return secret_key


_FERNET_SECRET_KEY = Fernet.generate_key()


def encrypt_object_to_base64(obj):
    try:
        dumped = json.dumps(obj).encode()
        compressed_dumped = zlib.compress(dumped)
        data = Fernet(_FERNET_SECRET_KEY).encrypt(compressed_dumped)
        return base64.b64encode(data).decode()
    except Exception:  # pylint: disable=broad-except
        return None


def decrypt_base64_to_object(text):
    try:
        base64_decoded = base64.b64decode(text.encode())
        compressed_dumped = Fernet(_FERNET_SECRET_KEY).decrypt(base64_decoded)
        dumped = zlib.decompress(compressed_dumped)
        return json.loads(dumped.decode())
    except Exception:  # pylint: disable=broad-except
        return None


def _now_timestamp() -> int:
    return int(datetime.now().timestamp())


def encrypt_local_file_ref(path: str, filename: str, max_expired: int = 300):
    return encrypt_object_to_base64({
            'path': path,
            'filename': filename,
            'timestamp': _now_timestamp(),
            'max_expired': max_expired,
    })


def decrypt_local_file_ref(text):
    msg = decrypt_base64_to_object(text)
    if msg is None:
        return False, None, None

    path = msg.get('path')
    filename = msg.get('filename')
    timestamp = msg.get('timestamp')
    max_expired = msg.get('max_expired')
    if not all((path, filename, timestamp, max_expired)):
        return False, None, None

    if _now_timestamp() - timestamp >= max_expired:
        return False, None, None

    return True, path, filename


# https://github.com/pypa/pip/blob/716afdb4cf4783ba2f610c2010aa76c4ffdb22e7/src/pip/_internal/utils/filetypes.py
_ARCHIVE_EXTENSIONS = {
        '.zip',
        '.whl',
        '.tar.bz2',
        '.tbz',
        '.tar.gz',
        '.tgz',
        '.tar',
        '.tar.xz',
        '.txz',
        '.tlz',
        '.tar.lz',
        '.tar.lzma',
}

_ARCHIVE_EXTENSION_LENGTHS = sorted(set(map(len, _ARCHIVE_EXTENSIONS)), reverse=True)


def split_package_ext(filename):
    for ext_len in _ARCHIVE_EXTENSION_LENGTHS:
        if len(filename) > ext_len and filename[-ext_len:] in _ARCHIVE_EXTENSIONS:
            package = filename[:-ext_len]
            ext = filename[1 - ext_len:]  # Remove the leading dot.
            return package, ext
    return '', None



LOGIN_HTML = '''
<html>

<head>
  <meta charset="UTF-8">
  <title>Login</title>
  <link href="https://fonts.googleapis.com/icon?family=Material+Icons" rel="stylesheet">
  <style>
    @import url('https://fonts.googleapis.com/css?family=Raleway');

    body {
      margin: 0;
      padding: 0;
      font-family: 'Raleway', sans-serif;
      color: #F2F2F2;
    }

    #container-login {
      background-color: #1D1F20;
      position: relative;
      top: 20%;
      margin: auto;
      width: 400px;
      height: 275px;
      border-radius: 0.35em;
      box-shadow: 0 3px 10px 0 rgba(0, 0, 0, 0.2);
      text-align: center;
    }

    #title {
      position: relative;
      background-color: #1A1C1D;
      width: 100%;
      padding: 20px 0px;
      border-radius: 0.35em;
      font-size: 22px;
      border-bottom: 1px solid rgba(255, 255, 255, 0.05);
    }

    .lock {
      position: relative;
      top: 2px;
    }

    .input {
      margin: auto;
      width: 300px;
      border-radius: 4px;
      background-color: #373b3d;
      padding: 8px 0px;
      margin-top: 15px;
    }

    .input-addon {
      float: left;
      background-color: #373b3d;
      border: 1px solid #373b3d;
      padding: 4px 8px;
      border-right: 1px solid rgba(255, 255, 255, 0.05);
    }

    input[type=text] {
      font-size: 16px;
      width: 215px;
      color: #949494;
      margin: 0;
      background-color: #373b3d;
      border: 1px solid #373b3d;
      padding: 6px 0px;
      border-radius: 3px;
    }

    input[type=text]:focus {
      border: 1px solid #373b3d;
    }

    input[type=password] {
      font-size: 16px;
      width: 215px;
      color: #949494;
      margin: 0;
      background-color: #373b3d;
      border: 1px solid #373b3d;
      padding: 6px 0px;
      border-radius: 3px;
    }

    input[type=password]:focus {
      border: 1px solid #373b3d;
    }

    *:focus {
      outline: none;
    }

    input[type=submit] {
      font-size: 14px;
      margin-top: 20px;
      padding: 10px 35px;
      background: #373E4A;
      color: #C1C3C6;
      font-weight: bold;
      border: 0 none;
      cursor: pointer;
      border-radius: 3px;
    }

    .clearfix {
      clear: both;
    }

    body {
      background-color: #303641;
    }
  </style>
</head>

<body>
  <div id="container-login">
    <div id="title">
      <i class="material-icons lock">lock</i> Login </div>
    <form action='/login/' method='POST'>
      <div class="input">
        <div class="input-addon">
          <i class="material-icons">face</i>
        </div>
        <input name="pkg_repo_name" placeholder="Package Repository Name" type="text" required class="validate" autocomplete="off">
      </div>
      <div class="clearfix"></div>
      <div class="input">
        <div class="input-addon">
          <i class="material-icons">vpn_key</i>
        </div>
        <input name="pkg_repo_secret_raw" placeholder="Package Repository Secret" type="password" required class="validate" autocomplete="off">
      </div>
      <input type="submit" value="Log In" />
    </form>
  </div>
</body>

</html>
'''



from dataclasses import dataclass
from os.path import join
from typing import Any, Optional, Tuple
import uuid
import logging

import fire
from flask import Flask, current_app, redirect, request, session, send_file
from flask_login import LoginManager, UserMixin, current_user, login_required
import waitress
from paste.translogger import TransLogger

from private_pypi.backend import PkgRepoSecret
from private_pypi.workflow import (
        WorkflowStat,
        initialize_workflow,
        workflow_api_redirect_package_download_url,
        workflow_api_simple,
        workflow_api_simple_distrib,
        workflow_api_upload_package,
)
from private_pypi.utils import get_secret_key, decrypt_local_file_ref, split_package_ext
from private_pypi.web import LOGIN_HTML

app = Flask(__name__)  # pylint: disable=invalid-name
app.secret_key = get_secret_key()

login_manager = LoginManager()  # pylint: disable=invalid-name
login_manager.init_app(app)
login_manager.login_view = 'login'


@dataclass
class MockUser(UserMixin):
    pkg_repo_name: str
    pkg_repo_secret_raw: str


SESSION_KEY_PKG_REPO_NAME = '_private_pypi_pkg_repo_name'
SESSION_KEY_PKG_REPO_SECRET_RAW = '_private_pypi_pkg_repo_secret_raw'


# https://github.com/maxcountryman/flask-login/blob/d4fa75305fdfb73bb55386d95bc09664bca8f902/flask_login/login_manager.py#L330-L331
@login_manager.request_loader
def load_user_from_request(_):
    # UA of pip and poetry:
    # https://github.com/pypa/pip/blob/7420629800b10d117d3af3b668dbe99b475fcbc0/src/pip/_internal/network/session.py#L99
    # https://github.com/python-poetry/poetry/blob/5050362f0b4c41d4637dcaa74eb2ba188bd858a9/get-poetry.py#L906
    if 'python' not in request.user_agent.string.lower():
        # Is browser.
        pkg_repo_name = session.get(SESSION_KEY_PKG_REPO_NAME)
        pkg_repo_secret_raw = session.get(SESSION_KEY_PKG_REPO_SECRET_RAW)
        if not pkg_repo_name or not pkg_repo_secret_raw:
            return None
        return MockUser(pkg_repo_name=pkg_repo_name, pkg_repo_secret_raw=pkg_repo_secret_raw)

    else:
        # In CLI, always returns a mock user to defer authentication.
        if request.authorization is None:
            username = ''
            password = ''
        else:
            username = request.authorization['username']
            password = request.authorization['password']
        return MockUser(pkg_repo_name=username, pkg_repo_secret_raw=password)


@app.route('/login/', methods=['GET', 'POST'])
def login():
    if request.method == 'GET':
        return LOGIN_HTML

    pkg_repo_name = request.form.get('pkg_repo_name')
    pkg_repo_secret_raw = request.form.get('pkg_repo_secret_raw')

    if not pkg_repo_name or not pkg_repo_secret_raw:
        return 'Repository name or secret is empty, please refresh this page and submit.', 401

    session[SESSION_KEY_PKG_REPO_NAME] = pkg_repo_name
    session[SESSION_KEY_PKG_REPO_SECRET_RAW] = pkg_repo_secret_raw

    return redirect(request.args.get('next') or '/simple/')


@app.route('/logout/', methods=['GET'])
def logout():
    session[SESSION_KEY_PKG_REPO_NAME] = None
    session[SESSION_KEY_PKG_REPO_SECRET_RAW] = None
    return redirect('/login/')


def load_name_from_request() -> str:
    return current_user.pkg_repo_name.lower()


def load_secret_from_request(wstat: WorkflowStat) -> Tuple[Optional[PkgRepoSecret], str]:
    name = load_name_from_request()
    if not name:
        return None, 'Empty package pepository name.'

    pkg_repo_config = wstat.name_to_pkg_repo_config.get(name)
    if pkg_repo_config is None:
        return None, f'Package repository name "{name}" not exists.'

    if not current_user.pkg_repo_secret_raw:
        return None, f'Secret of the package repository "{name}" is empty.'

    pkg_repo_secret = wstat.backend_instance_manager.create_pkg_repo_secret(
            name=name,
            type=pkg_repo_config.type,
            raw=current_user.pkg_repo_secret_raw,
    )
    return pkg_repo_secret, ''


@app.route('/simple/', methods=['GET'])
@login_required
def pep503_api_simple():
    pkg_repo_secret, err_msg = load_secret_from_request(current_app.workflow_stat)
    if pkg_repo_secret is None:
        return err_msg, 401

    name = load_name_from_request()
    body, status_code = workflow_api_simple(current_app.workflow_stat, name, pkg_repo_secret)
    return body, status_code


@app.route('/simple/<distrib>/', methods=['GET'])
@login_required
def pep503_api_simple_distrib(distrib):
    pkg_repo_secret, err_msg = load_secret_from_request(current_app.workflow_stat)
    if pkg_repo_secret is None:
        return err_msg, 401

    name = load_name_from_request()
    body, status_code = workflow_api_simple_distrib(
            current_app.workflow_stat,
            name,
            pkg_repo_secret,
            distrib,
    )

    if status_code == 404 and app.config['EXTRA_INDEX_URL'] != '/':
        # Redirect to extra index if not found.
        return redirect(app.config['EXTRA_INDEX_URL'] + f'/{distrib}/')

    return body, status_code


@app.route('/simple/<distrib>/<filename>', methods=['GET'])
@login_required
def pep503_api_redirect_package_download_url(distrib, filename):
    package, ext = split_package_ext(filename)
    if not ext:
        return 'Extension not supported.', 404

    pkg_repo_secret, err_msg = load_secret_from_request(current_app.workflow_stat)
    if pkg_repo_secret is None:
        return err_msg, 401

    name = load_name_from_request()
    auth_url, err_msg, status_code = workflow_api_redirect_package_download_url(
            current_app.workflow_stat,
            name,
            pkg_repo_secret,
            distrib,
            package,
            ext,
    )
    if auth_url is None:
        return err_msg, status_code
    return redirect(auth_url)


@app.route('/local_file/<encrypted_ref>', methods=['GET'])
def local_file(encrypted_ref):
    passed, path, filename = decrypt_local_file_ref(encrypted_ref)
    if not passed:
        return 'Expired or invalid encrypted_ref.', 401

    rsp = send_file(path, as_attachment=True, attachment_filename=filename)
    rsp.direct_passthrough = False
    return rsp


# https://warehouse.pypa.io/api-reference/legacy/#upload-api
@app.route('/simple/', methods=['POST'])
@login_required
def legacy_api_upload_package():  # pylint: disable=too-many-return-statements
    cache_folder = current_app.workflow_stat.root_local_paths.cache

    pkg_repo_secret, err_msg = load_secret_from_request(current_app.workflow_stat)
    if pkg_repo_secret is None:
        return err_msg, 401

    name = load_name_from_request()

    if 'multipart/form-data' not in request.content_type:
        return 'Please post content-type=multipart/form-data.', 405
    if 'content' not in request.files:
        return 'File not found.', 405

    # Save to the cache folder.
    content_file = request.files['content']
    cache_path = join(cache_folder, f'upload-{str(uuid.uuid1())}')
    content_file.save(cache_path)

    # Build meta.
    meta = {str(key): str(val) for key, val in request.form.items()}

    # Upload file.
    body, status_code = workflow_api_upload_package(
            current_app.workflow_stat,
            name,
            pkg_repo_secret,
            content_file.filename,
            meta,
            cache_path,
    )
    return body, status_code


def run_server(
        config: str,
        root: str,
        admin_secret: Optional[str] = None,
        auth_read_expires: int = 3600,
        auth_write_expires: int = 300,
        extra_index_url: str = 'https://pypi.org/simple/',
        debug: bool = False,
        host: str = 'localhost',
        port: int = 8888,
        **waitress_options: Any,
):
    # Make sure EXTRA_INDEX_URL ends with slash.
    # NOTE: EXTRA_INDEX_URL will be set as '/' if --extra_index_url=''.
    app.config['EXTRA_INDEX_URL'] = extra_index_url.rstrip('/') + '/'

    with app.app_context():
        # Init.
        current_app.workflow_stat = initialize_workflow(
                pkg_repo_config_file=config,
                admin_pkg_repo_secret_file=admin_secret,
                root_folder=root,
                auth_read_expires=auth_read_expires,
                auth_write_expires=auth_write_expires,
        )
        server_logging_path = join(
                current_app.workflow_stat.root_local_paths.log,
                'private_pypi_server.log',
        )

    # Setup logging.
    logging.basicConfig(level=logging.INFO)
    logging.getLogger('filelock').setLevel(logging.WARNING)
    logging.getLogger('apscheduler').setLevel(logging.WARNING)
    logging.getLogger().addHandler(logging.FileHandler(server_logging_path))

    if debug:
        if waitress_options:
            raise RuntimeError(
                    f'--waitress_options={waitress_options} should be empty in debug mode.')

        def print_request():
            print('==== REQUEST URL ====')
            print(request.url)
            print('==== REQUEST HEADERS ====')
            print(str(request.headers).strip())
            print()

        def print_response(response):
            print('==== RESPONSE HEADERS ====')
            print(str(response.headers).strip())
            print('==== RESPONSE DATA ====')
            print(response.get_data())
            print()
            return response

        app.before_request(print_request)
        app.after_request(print_response)

        app.run(
                host=host,
                port=port,
                load_dotenv=False,
                # Must be threaded for the current design.
                threaded=True,
        )

    else:
        print(f'waitress.serve host={host}, port={port}, waitress_options={waitress_options}')

        # https://docs.pylonsproject.org/projects/waitress/en/stable/logging.html
        trans_logger_wrapped_app = TransLogger(app, setup_console_handler=False)

        # https://docs.pylonsproject.org/projects/waitress/en/stable/arguments.html
        waitress.serve(
                trans_logger_wrapped_app,
                host=host,
                port=port,
                **waitress_options,
        )


run_server_cli = lambda: fire.Fire(run_server)  # pylint: disable=invalid-name



import atexit
from collections import defaultdict
import contextlib
from dataclasses import dataclass
from datetime import datetime
import logging
import os
from os.path import abspath, exists, getmtime, getsize, join
import subprocess
import socket
import threading
import traceback
from typing import DefaultDict, Dict, Generic, List, Optional, Tuple, TypeVar

from filelock import FileLock
from jinja2 import Template
import psutil
import redis_server
from apscheduler.schedulers.background import BackgroundScheduler

from private_pypi.backend import (
        DownloadIndexStatus,
        LocalPaths,
        PkgRef,
        PkgRepo,
        PkgRepoConfig,
        PkgRepoSecret,
        PkgRepoIndex,
        UploadPackageStatus,
        BackendInstanceManager,
)
from private_pypi.job import dynamic_dramatiq
from private_pypi.utils import locked_copy_file

SHST = TypeVar('SHST')


class SecretHashedStorage(Generic[SHST]):

    def __init__(self) -> None:
        self._hash_to_item: Dict[str, SHST] = {}

    def has_item(self, secret: PkgRepoSecret) -> bool:
        return secret.secret_hash() in self._hash_to_item

    def set_item(self, secret: PkgRepoSecret, item: SHST) -> None:
        self._hash_to_item[secret.secret_hash()] = item

    def get_item(self, secret: PkgRepoSecret) -> SHST:
        return self._hash_to_item[secret.secret_hash()]


def get_mtime_size(path: str) -> Tuple[datetime, int]:
    mtime = datetime.fromtimestamp(getmtime(path))
    size = getsize(path)
    return mtime, size


@dataclass
class WorkflowStat:
    # Backend reflection.
    backend_instance_manager: BackendInstanceManager

    # Package repository configs.
    name_to_pkg_repo_config: Dict[str, PkgRepoConfig]

    # Local paths.
    root_local_paths: LocalPaths
    name_to_local_paths: Dict[str, LocalPaths]

    # Admin secrets for index synchronization.
    name_to_admin_pkg_repo_secret: Optional[Dict[str, PkgRepoSecret]]

    # Package index paths [(<lock-path>, <toml-path>)...]
    name_to_index_paths: Dict[str, Tuple[str, str]]
    name_to_index_mtime_size: Dict[str, Tuple[datetime, int]]
    name_to_pkg_repo_index: Dict[str, PkgRepoIndex]

    # Package repositories guarded by threading locks.
    auth_read_expires: int
    auth_write_expires: int
    pkg_repo_global_lock: threading.Lock
    name_to_pkg_repo_lock_shstg: DefaultDict[str, SecretHashedStorage[threading.RLock]]
    name_to_pkg_repo_shstg: DefaultDict[str, SecretHashedStorage[PkgRepo]]
    # Read/write last succeeded authentication datetime.
    name_to_pkg_repo_read_mtime_shstg: DefaultDict[str, SecretHashedStorage[datetime]]
    name_to_pkg_repo_write_mtime_shstg: DefaultDict[str, SecretHashedStorage[datetime]]

    scheduler: BackgroundScheduler


def build_workflow_stat(
        pkg_repo_config_file: str,
        admin_pkg_repo_secret_file: Optional[str],
        root_folder: str,
        auth_read_expires: int,
        auth_write_expires: int,
) -> WorkflowStat:
    backend_instance_manager = BackendInstanceManager()

    # Config.
    if not exists(pkg_repo_config_file):
        raise FileNotFoundError(f'pkg_repo_config_file={pkg_repo_config_file} not exists.')

    name_to_pkg_repo_config = backend_instance_manager.load_pkg_repo_configs(pkg_repo_config_file)

    # Admin secret.
    name_to_admin_pkg_repo_secret = None
    if admin_pkg_repo_secret_file is not None:
        name_to_admin_pkg_repo_secret = \
                backend_instance_manager.load_pkg_repo_secrets(admin_pkg_repo_secret_file)

    # Root folders.
    root_local_paths = LocalPaths(
            index=join(root_folder, 'index'),
            log=join(root_folder, 'log'),
            lock=join(root_folder, 'lock'),
            job=join(root_folder, 'job'),
            cache=join(root_folder, 'cache'),
    )
    root_local_paths.makedirs()

    name_to_local_paths = {}
    name_to_index_paths = {}
    for pkg_repo_config in name_to_pkg_repo_config.values():
        name = pkg_repo_config.name

        # Create isolated folders for each package repository.
        name_to_local_paths[name] = LocalPaths(
                index=join(root_local_paths.index, name),
                log=join(root_local_paths.log, name),
                lock=join(root_local_paths.lock, name),
                job=join(root_local_paths.job, name),
                cache=join(root_local_paths.cache, name),
        )
        name_to_local_paths[name].makedirs()

        # Index paths of (index_lock, index).
        name_to_index_paths[name] = (
                join(name_to_local_paths[name].lock, f'index.toml.lock'),
                join(name_to_local_paths[name].index, f'index.toml'),
        )

    # Build WorkflowStat.
    wstat = WorkflowStat(
            backend_instance_manager=backend_instance_manager,
            name_to_pkg_repo_config=name_to_pkg_repo_config,
            root_local_paths=root_local_paths,
            name_to_local_paths=name_to_local_paths,
            name_to_admin_pkg_repo_secret=name_to_admin_pkg_repo_secret,
            name_to_index_paths=name_to_index_paths,
            name_to_index_mtime_size={},  # Will setup later.
            name_to_pkg_repo_index={},  # Will setup later.
            auth_read_expires=auth_read_expires,
            auth_write_expires=auth_write_expires,
            pkg_repo_global_lock=threading.Lock(),
            name_to_pkg_repo_lock_shstg=defaultdict(SecretHashedStorage),
            name_to_pkg_repo_shstg=defaultdict(SecretHashedStorage),
            name_to_pkg_repo_read_mtime_shstg=defaultdict(SecretHashedStorage),
            name_to_pkg_repo_write_mtime_shstg=defaultdict(SecretHashedStorage),
            scheduler=BackgroundScheduler(),
    )

    if name_to_admin_pkg_repo_secret:
        passed, log = sync_local_index(wstat)
        if not passed:
            raise RuntimeError(log)

    # Index file signature (mtime, size) and instance.
    for pkg_repo_config in name_to_pkg_repo_config.values():
        _, index_path = name_to_index_paths[pkg_repo_config.name]
        if not exists(index_path):
            raise FileNotFoundError(
                    f'index file={index_path} for name={pkg_repo_config.name} not exists')

        wstat.name_to_index_mtime_size[pkg_repo_config.name] = get_mtime_size(index_path)
        wstat.name_to_pkg_repo_index[pkg_repo_config.name] = \
                PkgRepoIndex(wstat.backend_instance_manager.load_pkg_refs(index_path))

    return wstat


def sync_single_local_index(wstat: WorkflowStat, name: str) -> Tuple[bool, str]:
    pkg_repo_config = wstat.name_to_pkg_repo_config[name]

    pkg_repo_secret = wstat.name_to_admin_pkg_repo_secret.get(name)
    if pkg_repo_secret is None:
        return True, f'[WARN] secret of "{name}" is not provided, skip index sync.'

    index_lock_path, index_path = wstat.name_to_index_paths[name]
    timestamp = datetime.now().strftime('%Y%m%d%H%M%S%f')
    index_tmp_path = f'{index_path}.tmp.{timestamp}'

    try:
        pkg_repo = wstat.backend_instance_manager.create_pkg_repo(
                type=pkg_repo_config.type,
                config=pkg_repo_config,
                secret=pkg_repo_secret,
                local_paths=wstat.name_to_local_paths[name],
        )

        up_to_date = False
        if exists(index_path):
            # Make a copy to avoid holding the lock for a long time.
            locked_copy_file(index_lock_path, index_path, index_tmp_path, timeout=0.1)
            # Check.
            up_to_date = pkg_repo.local_index_is_up_to_date(index_tmp_path)

        if not up_to_date:
            # Sync.
            result = pkg_repo.download_index(index_tmp_path)
            if result.status != DownloadIndexStatus.SUCCEEDED:
                return False, f'[ERROR] "{name}" failed to download index:\n' + result.message
            # And replace.
            locked_copy_file(index_lock_path, index_tmp_path, index_path, timeout=0.1)

        # index_tmp_path should exists.
        try:
            os.remove(index_tmp_path)
        except IOError:
            pass

        return True, f'[PASS] "{name}" is up-to-date.'

    except Exception:  # pylint: disable=broad-except
        return False, f'[ERROR] traceback of "{name}":\n' + traceback.format_exc()


def sync_local_index(wstat: WorkflowStat) -> Tuple[bool, str]:
    if wstat.name_to_admin_pkg_repo_secret is None:
        return False, 'name_to_admin_pkg_repo_secret is None.'

    all_passed = []
    all_logs = []
    for name in wstat.name_to_pkg_repo_config:
        passed, log = sync_single_local_index(wstat, name)
        all_passed.append(passed)
        all_logs.append(log)

    return all(all_passed), '\n'.join(all_logs)


@dynamic_dramatiq.actor()
def sync_local_index_job(
        pkg_repo_config_file: str,
        admin_pkg_repo_secret_file: str,
        root_folder: str,
        name: str,
):
    wstat = build_workflow_stat(
            pkg_repo_config_file=pkg_repo_config_file,
            admin_pkg_repo_secret_file=admin_pkg_repo_secret_file,
            root_folder=root_folder,
            auth_read_expires=0,
            auth_write_expires=0,
    )

    # Setup logging.
    logging.basicConfig(level=logging.INFO)
    logging.getLogger('filelock').setLevel(logging.WARNING)
    logger = logging.getLogger()
    logging_path = join(wstat.name_to_local_paths[name].log, 'sync_local_index_job.log')
    logger.addHandler(logging.FileHandler(logging_path))

    # Sync.
    try:
        passed, log = sync_single_local_index(wstat, name)
        if passed:
            logger.info(log)
        else:
            logger.error(log)

    except Exception:  # pylint: disable=broad-except
        logger.error(traceback.format_exc())


def random_select_port() -> str:
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as _socket:
        _socket.bind(('localhost', 0))
        _socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        _, port = _socket.getsockname()
        return str(port)


def stop_all_children_processes():
    procs = psutil.Process().children()
    for proc in procs:
        proc.terminate()

    _, alive = psutil.wait_procs(procs, timeout=10)
    for proc in alive:
        proc.kill()


def initialize_task_worker():
    # All processes in the current process group will be terminated
    # with the lead process.
    os.setpgrp()
    atexit.register(stop_all_children_processes)

    # Run Redis.
    redis_port = random_select_port()
    pgid = os.getpgrp()
    subprocess.Popen(  # pylint: disable=subprocess-popen-preexec-fn
            [redis_server.REDIS_SERVER_PATH, '--port', redis_port],
            # Attach to the current process group.
            preexec_fn=lambda: os.setpgid(0, pgid),
            # Suppress stdout and stderr.
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
    )

    # Set broker.
    from dramatiq.brokers.redis import RedisBroker  # pylint: disable=import-outside-toplevel
    dynamic_dramatiq.set_broker(RedisBroker(host='localhost', port=redis_port))

    # Run worker.
    dramatiq_env = dict(os.environ)
    dramatiq_env['DYNAMIC_DRAMATIQ_REDIS_BROKER_PORT'] = redis_port
    subprocess.Popen(  # pylint: disable=subprocess-popen-preexec-fn
            ['dramatiq', 'private_pypi.job', '--processes', '1'],
            # Share env.
            env=dramatiq_env,
            # Attach to the current process group.
            preexec_fn=lambda: os.setpgid(0, pgid),
            # Suppress stdout and stderr.
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
    )


def initialize_workflow(
        pkg_repo_config_file: str,
        admin_pkg_repo_secret_file: Optional[str],
        root_folder: str,
        auth_read_expires: int,
        auth_write_expires: int,
) -> WorkflowStat:
    # Initialize workflow state.
    # NOTE: backend_instance_manager must be created before broker setup.
    wstat = build_workflow_stat(
            pkg_repo_config_file=pkg_repo_config_file,
            admin_pkg_repo_secret_file=admin_pkg_repo_secret_file,
            root_folder=abspath(root_folder),
            auth_read_expires=auth_read_expires,
            auth_write_expires=auth_write_expires,
    )

    # Initialize task queue related stuff.
    initialize_task_worker()

    # Schedule sync_local_index_job.
    for name, pkg_repo_config in wstat.name_to_pkg_repo_config.items():
        wstat.scheduler.add_job(
                sync_local_index_job,
                trigger='interval',
                kwargs={
                        'pkg_repo_config_file': pkg_repo_config_file,
                        'admin_pkg_repo_secret_file': admin_pkg_repo_secret_file,
                        'root_folder': root_folder,
                        'name': name,
                },
                seconds=pkg_repo_config.sync_index_interval,
        )
    wstat.scheduler.start()

    return wstat


def pkg_repo_is_expired(
        wstat: WorkflowStat,
        name: str,
        pkg_repo_lock: threading.RLock,
        pkg_repo_secret: PkgRepoSecret,
        check_auth_read: bool,
) -> bool:
    with pkg_repo_lock:
        pkg_repo_shstg = wstat.name_to_pkg_repo_shstg[name]

        if check_auth_read:
            pkg_repo_mtime_shstg = wstat.name_to_pkg_repo_read_mtime_shstg[name]
        else:
            pkg_repo_mtime_shstg = wstat.name_to_pkg_repo_write_mtime_shstg[name]

        if not pkg_repo_shstg.has_item(pkg_repo_secret):
            # Cannot find the instance.
            return True

        pkg_repo = pkg_repo_shstg.get_item(pkg_repo_secret)
        ready, _ = pkg_repo.ready()
        if not ready:
            # Error in the last request.
            return True

        if not pkg_repo_mtime_shstg.has_item(pkg_repo_secret):
            # Unknow.
            return True

        mtime = pkg_repo_mtime_shstg.get_item(pkg_repo_secret)
        auth_expires = wstat.auth_read_expires if check_auth_read else wstat.auth_write_expires
        if (datetime.now() - mtime).total_seconds() >= auth_expires:
            # Last passed but time expires.
            return True

        # Last passed and not expired.
        return False


def pkg_repo_secret_is_authenticated(
        wstat: WorkflowStat,
        name: str,
        pkg_repo_secret: PkgRepoSecret,
        check_auth_read: bool,
) -> Tuple[Optional[PkgRepo], str]:
    '''name has been validated.
    '''
    pkg_repo_config = wstat.name_to_pkg_repo_config[name]

    # Get package repository lock.
    with wstat.pkg_repo_global_lock:
        pkg_repo_lock_shstg = wstat.name_to_pkg_repo_lock_shstg[name]

        if not pkg_repo_lock_shstg.has_item(pkg_repo_secret):
            pkg_repo_lock_shstg.set_item(pkg_repo_secret, threading.RLock())
        pkg_repo_lock = pkg_repo_lock_shstg.get_item(pkg_repo_secret)

    # Prepare the package repository.
    with pkg_repo_lock:
        pkg_repo_shstg = wstat.name_to_pkg_repo_shstg[name]

        if check_auth_read:
            pkg_repo_mtime_shstg = wstat.name_to_pkg_repo_read_mtime_shstg[name]
        else:
            pkg_repo_mtime_shstg = wstat.name_to_pkg_repo_write_mtime_shstg[name]

        if pkg_repo_is_expired(
                wstat,
                name,
                pkg_repo_lock,
                pkg_repo_secret,
                check_auth_read,
        ):
            # Initialize.
            pkg_repo = wstat.backend_instance_manager.create_pkg_repo(
                    type=pkg_repo_config.type,
                    config=pkg_repo_config,
                    secret=pkg_repo_secret,
                    local_paths=wstat.name_to_local_paths[name],
            )

            ready, err_msg = pkg_repo.ready()
            if ready:
                auth_passed = pkg_repo.auth_read() if check_auth_read else pkg_repo.auth_write()
                if not auth_passed:
                    ready = False
                    err_msg = f'Auth error (readonly={check_auth_read})'
            else:
                err_msg = f'Auth setup error (readonly={check_auth_read})\n' + err_msg
            if not ready:
                return None, err_msg

            # Succeeded.
            pkg_repo_shstg.set_item(pkg_repo_secret, pkg_repo)
            pkg_repo_mtime_shstg.set_item(pkg_repo_secret, datetime.now())

        return pkg_repo_shstg.get_item(pkg_repo_secret), ''


def keep_pkg_repo_index_up_to_date(wstat: WorkflowStat, name: str) -> Tuple[bool, str]:
    pkg_repo_config = wstat.name_to_pkg_repo_config[name]
    index_lock_path, index_path = wstat.name_to_index_paths[name]

    try:
        # Timeout = 1.0 second should be enough
        # since the current design is only for small index file.
        with FileLock(index_lock_path, timeout=1.0):
            cur_mtime_size = get_mtime_size(index_path)
            last_mtime_size = wstat.name_to_index_mtime_size[name]

            if cur_mtime_size != last_mtime_size:
                # Index has been updated, reload.
                wstat.name_to_pkg_repo_index[pkg_repo_config.name] = \
                        PkgRepoIndex(wstat.backend_instance_manager.load_pkg_refs(index_path))

    except Exception:  # pylint: disable=broad-except
        return False, traceback.format_exc()

    return True, ''


def get_pkg_repo_index(wstat: WorkflowStat, name: str) -> Tuple[Optional[PkgRepoIndex], str]:
    index_lock_path, _ = wstat.name_to_index_paths[name]
    try:
        with FileLock(index_lock_path, timeout=1.0):
            return wstat.name_to_pkg_repo_index[name], ''

    except Exception:  # pylint: disable=broad-except
        return None, traceback.format_exc()


@dataclass
class LinkItem:
    href: str
    text: str


# PEP 503 -- Simple Repository API
# https://www.python.org/dev/peps/pep-0503/
PAGE_TEMPLATE = Template('''<!DOCTYPE html>
<html>
<head><title>{{ title }}</title></head>
<body>
<h1>{{ title }}</h1>

{% for link_item in link_items %}
    <a href="{{ link_item.href }}">{{ link_item.text }}</a>
    <br>
{% endfor %}

</body>
</html>
''')


def build_page_api_simple(pkg_repo_index: PkgRepoIndex) -> str:
    link_items = [
            LinkItem(href=f'{distrib}/', text=distrib)
            for distrib in pkg_repo_index.all_distributions
    ]
    return PAGE_TEMPLATE.render(
            title='Links for all distributions',
            link_items=link_items,
    )


def build_page_api_simple_distrib(distrib: str, pkg_refs: List[PkgRef]) -> str:
    link_items = []
    for pkg_ref in pkg_refs:
        link_items.append(
                LinkItem(
                        href=f'{pkg_ref.package}.{pkg_ref.ext}#sha256={pkg_ref.sha256}',
                        text=f'{pkg_ref.package}.{pkg_ref.ext}',
                ))
    return PAGE_TEMPLATE.render(
            title=f'Links for {distrib}',
            link_items=link_items,
    )


def workflow_get_pkg_repo_index(
        wstat: WorkflowStat,
        name: str,
        pkg_repo_secret: PkgRepoSecret,
        check_auth_read: bool = True,
) -> Tuple[Optional[PkgRepoIndex], str, int]:
    pkg_repo, err_msg = pkg_repo_secret_is_authenticated(
            wstat,
            name,
            pkg_repo_secret,
            check_auth_read=check_auth_read,
    )
    if pkg_repo is None:
        return None, err_msg, 401

    passed_index, err_msg = keep_pkg_repo_index_up_to_date(wstat, name)
    if not passed_index:
        return None, err_msg, 404

    pkg_repo_index, err_msg = get_pkg_repo_index(wstat, name)
    if pkg_repo_index is None:
        return None, err_msg, 404

    return pkg_repo_index, '', -1


def workflow_api_simple(
        wstat: WorkflowStat,
        name: str,
        pkg_repo_secret: PkgRepoSecret,
) -> Tuple[str, int]:
    pkg_repo_index, err_msg, status_code = workflow_get_pkg_repo_index(
            wstat,
            name,
            pkg_repo_secret,
    )
    if pkg_repo_index is None:
        return err_msg, status_code
    return build_page_api_simple(pkg_repo_index), 200


def workflow_api_simple_distrib(
        wstat: WorkflowStat,
        name: str,
        pkg_repo_secret: PkgRepoSecret,
        distrib: str,
) -> Tuple[str, int]:
    pkg_repo_index, err_msg, status_code = workflow_get_pkg_repo_index(
            wstat,
            name,
            pkg_repo_secret,
    )
    if pkg_repo_index is None:
        return err_msg, status_code

    pkg_refs = pkg_repo_index.get_pkg_refs(distrib)
    if not pkg_refs:
        return f'distrib={distrib} not found.', 404

    return build_page_api_simple_distrib(distrib, pkg_refs), 200


def workflow_api_redirect_package_download_url(
        wstat: WorkflowStat,
        name: str,
        pkg_repo_secret: PkgRepoSecret,
        distrib: str,
        package: str,
        ext: str,
) -> Tuple[Optional[str], str, int]:
    pkg_repo_index, err_msg, status_code = workflow_get_pkg_repo_index(
            wstat,
            name,
            pkg_repo_secret,
    )
    if pkg_repo_index is None:
        return None, err_msg, status_code

    pkg_ref = pkg_repo_index.get_single_pkg_ref(distrib, package)
    if pkg_ref is None:
        return None, f'Package "{distrib}, {package}.{ext}" not exists.', 404
    elif pkg_ref.ext != ext:
        return None, f'Package "{distrib}, {package}.{ext}" extention not match (query="{ext}")', 404

    try:
        auth_url = pkg_ref.auth_url(wstat.name_to_pkg_repo_config[name], pkg_repo_secret)
        return auth_url, '', -1

    except Exception:  # pylint: disable=broad-except
        return None, 'Failed to get resource.\n' + traceback.format_exc(), 401


def workflow_api_upload_package(
        wstat: WorkflowStat,
        name: str,
        pkg_repo_secret: PkgRepoSecret,
        filename: str,
        meta: Dict[str, str],
        path: str,
) -> Tuple[str, int]:
    pkg_repo, err_msg = pkg_repo_secret_is_authenticated(
            wstat,
            name,
            pkg_repo_secret,
            check_auth_read=False,
    )
    if pkg_repo is None:
        return err_msg, 401

    result = pkg_repo.upload_package(filename, meta, path)
    if result.status == UploadPackageStatus.FAILED:
        status_code = 401
    elif result.status == UploadPackageStatus.SUCCEEDED:
        status_code = 200
    else:
        raise ValueError('Invalid UploadPackageStatus')

    return result.message, status_code



from private_pypi.backend import BackendRegistration
from private_pypi_backends.file_system.impl import (
        FILE_SYSTEM_TYPE,
        FileSystemConfig,
        FileSystemSecret,
        FileSystemPkgRepo,
        FileSystemPkgRef,
)


class FileSystemRegistration(BackendRegistration):
    type = FILE_SYSTEM_TYPE
    pkg_repo_config_cls = FileSystemConfig
    pkg_repo_secret_cls = FileSystemSecret
    pkg_repo_cls = FileSystemPkgRepo
    pkg_ref_cls = FileSystemPkgRef



from dataclasses import dataclass
import hashlib
from os import makedirs
from os.path import isdir, join, exists
import pathlib
import shutil
from typing import Dict, List, Tuple
import traceback

from filelock import FileLock
import shortuuid

from private_pypi.backend import (
        PkgRef,
        PkgRepo,
        PkgRepoConfig,
        PkgRepoSecret,
        UploadPackageStatus,
        UploadPackageResult,
        UploadPackageContext,
        UploadIndexStatus,
        UploadIndexResult,
        DownloadIndexStatus,
        DownloadIndexResult,
        BackendInstanceManager,
)
from private_pypi.utils import (
        write_toml,
        read_toml,
        git_hash_sha,
        encrypt_local_file_ref,
        split_package_ext,
)

FILE_SYSTEM_TYPE = 'file_system'


class FileSystemConfig(PkgRepoConfig):
    # Override.
    type: str = FILE_SYSTEM_TYPE
    max_file_bytes: int = 5 * 1024**3 - 1
    # File system specific.
    read_secret: str = ''
    write_secret: str = ''


class FileSystemSecret(PkgRepoSecret):
    # Override.
    type: str = FILE_SYSTEM_TYPE

    @property
    def token(self) -> str:
        return self.raw

    def secret_hash(self) -> str:
        sha256_algo = hashlib.sha256()
        sha256_algo.update(self.raw.encode())
        return f'fs-{sha256_algo.hexdigest()}'


class FileSystemPkgRef(PkgRef):
    # Override.
    type: str = FILE_SYSTEM_TYPE
    # File system specific.
    package_path: str

    def auth_url(self, config: FileSystemConfig, secret: FileSystemSecret) -> str:
        encrypted_ref = encrypt_local_file_ref(self.package_path, f'{self.package}.{self.ext}')
        if not encrypted_ref:
            raise RuntimeError(f'encrypt_local_file_ref failed ({self.package_path})')

        return f'/local_file/{encrypted_ref}#sha256={self.sha256}'


@dataclass
class FileSystemPkgRepoPrivateFields:
    ready: bool
    err_msg: str


LOCK_TIMEOUT = 0.5
META_SUFFIX = '.meta'


class FileSystemPkgRepo(PkgRepo):
    # Override.
    type: str = FILE_SYSTEM_TYPE
    # File system specific.
    config: FileSystemConfig
    secret: FileSystemSecret

    __slots__ = ('_private_fields',)

    @property
    def _pvt(self) -> FileSystemPkgRepoPrivateFields:
        return object.__getattribute__(self, '_private_fields')

    def __init__(self, **data):
        super().__init__(**data)
        object.__setattr__(
                self,
                '_private_fields',
                FileSystemPkgRepoPrivateFields(ready=True, err_msg=''),
        )

        if not isdir(self.local_paths.cache):
            self._pvt.ready = False
            self._pvt.err_msg = 'Cache path not exists'

        makedirs(self._storage_path, exist_ok=True)

    def record_error(self, error_message: str) -> None:
        self._pvt.ready = False
        self._pvt.err_msg = error_message

    def ready(self) -> Tuple[bool, str]:
        return self._pvt.ready, self._pvt.err_msg

    def auth_read(self) -> bool:
        return self.secret.token in (self.config.read_secret, self.config.write_secret)

    def auth_write(self) -> bool:
        return self.secret.token == self.config.write_secret

    def _package_lock_path(self, filename: str) -> str:
        return join(self.local_paths.lock, filename + '.lock')

    @property
    def _storage_path(self) -> str:
        return join(self.local_paths.cache, 'storage')

    @property
    def _index_path(self) -> str:
        return join(self._storage_path, 'index.toml')

    @property
    def _index_lock_path(self) -> str:
        return join(self.local_paths.lock, 'index.toml.lock')

    def _distrib_path(self, distrib: str) -> str:
        path = join(self._storage_path, distrib)
        makedirs(path, exist_ok=True)
        return path

    def _package_path(self, distrib: str, filename: str) -> str:
        distrib_path = self._distrib_path(distrib)
        return join(distrib_path, filename)

    def _package_meta_path(self, distrib: str, filename: str) -> str:
        distrib_path = self._distrib_path(distrib)
        return join(distrib_path, filename + META_SUFFIX)

    def _upload_package(self, ctx: UploadPackageContext):
        try:
            with FileLock(self._package_lock_path(ctx.filename), timeout=LOCK_TIMEOUT):
                pkg_path = self._package_path(ctx.meta_distrib, ctx.filename)
                pkg_meta_path = self._package_meta_path(ctx.meta_distrib, ctx.filename)

                if exists(pkg_path) and exists(pkg_meta_path):
                    ctx.failed = True
                    ctx.message = 'Package exists.'
                    return

                # Save package & meta.
                shutil.copyfile(ctx.path, pkg_path)
                write_toml(pkg_meta_path, ctx.meta)

            # Update index.
            pkg_refs = self.collect_all_published_packages()
            tmp_index_path = join(self.local_paths.cache, f'tmp_index_{shortuuid.uuid()}.toml')
            BackendInstanceManager.dump_pkg_refs(tmp_index_path, pkg_refs)
            self.upload_index(tmp_index_path)

        except TimeoutError:
            ctx.failed = True
            ctx.message = '_upload_package: Lock acquire timeout.'

        except Exception:  # pylint: disable=broad-except
            ctx.failed = True
            ctx.message = '_upload_package:\n' + traceback.format_exc()

    def upload_package(self, filename: str, meta: Dict[str, str], path: str) -> UploadPackageResult:
        ctx = UploadPackageContext(filename=filename, meta=meta, path=path)

        for action in (
                lambda _: None,  # Validate the context initialization.
                self._upload_package,
        ):
            action(ctx)
            if ctx.failed:
                break

        status = UploadPackageStatus.SUCCEEDED if not ctx.failed else UploadPackageStatus.FAILED
        return UploadPackageResult(status=status, message=ctx.message)

    def collect_all_published_packages(self) -> List[PkgRef]:
        pkg_refs = []

        storage_path = pathlib.Path(self._storage_path)
        for distrib_path in storage_path.iterdir():
            if not distrib_path.is_dir():
                continue

            filename_to_package = {}
            filename_to_meta = {}

            for child in distrib_path.iterdir():
                if child.name.endswith(META_SUFFIX):
                    filename = child.name[:-len(META_SUFFIX)]
                    filename_to_meta[filename] = str(child)
                else:
                    filename = child.name
                    filename_to_package[filename] = str(child)

            for filename in sorted(set(filename_to_package) & set(filename_to_meta)):
                meta = read_toml(filename_to_meta[filename])

                distrib = meta.get('distrib')
                sha256 = meta.get('sha256')
                if not distrib or not sha256:
                    continue

                package, ext = split_package_ext(filename)
                if not ext:
                    continue

                pkg_refs.append(
                        FileSystemPkgRef(
                                distrib=distrib,
                                package=package,
                                ext=ext,
                                sha256=sha256,
                                meta=meta,
                                package_path=filename_to_package[filename],
                        ))

        return pkg_refs

    def local_index_is_up_to_date(self, path: str) -> bool:
        try:
            with FileLock(self._index_lock_path, timeout=LOCK_TIMEOUT):
                return exists(self._index_path) \
                        and git_hash_sha(path) == git_hash_sha(self._index_path)
        except Exception:  # pylint: disable=broad-except
            return False

    def upload_index(self, path: str) -> UploadIndexResult:
        try:
            with FileLock(self._index_lock_path, timeout=LOCK_TIMEOUT):
                shutil.copyfile(path, self._index_path)
            return UploadIndexResult(status=UploadIndexStatus.SUCCEEDED)

        except TimeoutError:
            return UploadIndexResult(
                    status=UploadIndexStatus.FAILED,
                    message='upload_index: Lock acquire timeout.',
            )

        except Exception:  # pylint: disable=broad-except
            return UploadIndexResult(
                    status=UploadIndexStatus.FAILED,
                    message='upload_index:\n' + traceback.format_exc(),
            )

    def download_index(self, path: str) -> DownloadIndexResult:
        try:
            with FileLock(self._index_lock_path, timeout=LOCK_TIMEOUT):
                if not exists(self._index_path):
                    # Initialize index file.
                    BackendInstanceManager.dump_pkg_refs(self._index_path, [])
                shutil.copyfile(self._index_path, path)
            return DownloadIndexResult(status=DownloadIndexStatus.SUCCEEDED)

        except TimeoutError:
            return DownloadIndexResult(
                    status=DownloadIndexStatus.FAILED,
                    message='download_index: Lock acquire timeout.',
            )

        except Exception:  # pylint: disable=broad-except
            return DownloadIndexResult(
                    status=DownloadIndexStatus.FAILED,
                    message='download_index:\n' + traceback.format_exc(),
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['private_pypi',
 'private_pypi_backends',
 'private_pypi_backends.file_system',
 'private_pypi_testkit']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.1,<3.0.0',
 'Paste>=3.4.0,<4.0.0',
 'apscheduler>=3.6.3,<4.0.0',
 'cryptography>=2.8,<3.0',
 'dramatiq[redis]>=1.8.1,<2.0.0',
 'filelock>=3.0.12,<4.0.0',
 'fire>=0.2.1,<0.3.0',
 'flask-login>=0.5.0,<0.6.0',
 'flask>=1.1.1,<2.0.0',
 'psutil>=5.6.7,<6.0.0',
 'pydantic>=1.4,<2.0',
 'redis-server>=5.0.7,<6.0.0',
 'shortuuid>=0.5.0,<0.6.0',
 'toml>=0.10.0,<0.11.0',
 'waitress>=1.4.3,<2.0.0']

entry_points = \
{'console_scripts': ['private_pypi_server = '
                     'private_pypi.server:run_server_cli']}

setup_kwargs = {
    'name': 'private-pypi-core',
    'version': '0.1.3a2',
    'description': 'A private PyPI server powered by flexible backends.',
    'long_description': "# private-pypi-core\n\n## CLI\n\n`private_pypi_server`:\n\n```txt\nRun the private-pypi server.\n\nSYNOPSIS\n    private_pypi_server CONFIG ROOT <flags>\n\nPOSITIONAL ARGUMENTS\n    CONFIG (str):\n        Path to the package repositories config.\n    ROOT (str):\n        Path to the root folder.\n\nFLAGS\n    --admin_secret (Optional[str]):\n        Path to the admin secrets config with read/write permission.\n        This field is required for local index synchronization.\n        Defaults to None.\n    --auth_read_expires (int):\n        The expiration time (in seconds) for read authentication.\n        Defaults to 3600.\n    --auth_write_expires (int):\n        The expiration time (in seconds) for write authentication.\n        Defaults to 300.\n    --extra_index_url (str):\n        Extra index url for redirection in case package not found.\n        If set to empty string explicitly redirection will be suppressed.\n        Defaults to 'https://pypi.org/simple/'.\n    --debug (bool):\n        Enable debug mode.\n        Defaults to False.\n    --host (str):\n        The interface to bind to.\n        Defaults to 'localhost'.\n    --port (int):\n        The port to bind to.\n        Defaults to 8080.\n    **waitress_options (Dict[str, Any]):\n        Optional arguments that `waitress.serve` takes.\n        Details in https://docs.pylonsproject.org/projects/waitress/en/stable/arguments.html.\n        Defaults to {}.\n```\n",
    'author': 'huntzhan',
    'author_email': 'huntzhan.dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/private-pypi/private-pypi-core',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import abstractmethod
from dataclasses import dataclass
import os
from typing import Optional, Tuple
import inspect

import pytest
import shortuuid

from private_pypi.backend import (
        BackendInstanceManager,
        PkgRepoConfig,
        PkgRepoSecret,
)
from private_pypi.workflow import (
        WorkflowStat,
        initialize_task_worker,
        build_workflow_stat,
        sync_local_index,
        workflow_api_simple_distrib,
        workflow_api_upload_package,
)


@dataclass
class RepoInfoForTest:
    pkg_repo_config_file: str
    admin_pkg_repo_secret_file: str
    root_folder: str
    name: str
    wstat: WorkflowStat
    read_secret: PkgRepoSecret
    write_secret: PkgRepoSecret


def create_random_file(path, size):
    with open(path, 'wb') as fout:
        fout.write(os.urandom(size))
    return path


def test_upload_with_write_secret(session_repo, tmpdir, update_repo_index):
    distrib = shortuuid.uuid()
    filename = f'{distrib}-1.0-py3-none-any.whl'
    path = str(tmpdir.join(filename))
    create_random_file(path, 128)

    # Upload.
    _, status_code = workflow_api_upload_package(
            wstat=session_repo.wstat,
            name=session_repo.name,
            pkg_repo_secret=session_repo.write_secret,
            filename=filename,
            meta={'name': distrib},
            path=path,
    )
    assert status_code == 200

    # Update index in remote.
    assert update_repo_index(session_repo)

    # Sync local index.
    passed, _ = sync_local_index(session_repo.wstat)
    assert passed

    # Check if package exists.
    _, status_code = workflow_api_simple_distrib(
            wstat=session_repo.wstat,
            name=session_repo.name,
            pkg_repo_secret=session_repo.write_secret,
            distrib=distrib,
    )
    assert status_code == 200


def test_upload_with_read_secret(session_repo, tmpdir):
    distrib = shortuuid.uuid()
    filename = f'{distrib}-1.0-py3-none-any.whl'
    path = str(tmpdir.join(filename))
    create_random_file(path, 128)

    _, status_code = workflow_api_upload_package(
            wstat=session_repo.wstat,
            name=session_repo.name,
            pkg_repo_secret=session_repo.read_secret,
            filename=filename,
            meta={'name': distrib},
            path=path,
    )
    assert status_code == 401


class TestKit:

    @classmethod
    @abstractmethod
    def setup_pkg_repo(cls) -> Tuple[PkgRepoConfig, PkgRepoSecret, PkgRepoSecret]:
        pass

    @classmethod
    @abstractmethod
    def update_repo_index(cls, repo: RepoInfoForTest) -> bool:
        return False

    @classmethod
    def pytest_injection(cls):
        _caller_frame = inspect.currentframe().f_back

        def inject_to_caller(func):
            caller_globals = _caller_frame.f_globals
            caller_globals[func.__name__] = func
            return func

        def _create_repo_for_test(create_tmpdir):
            config_folder = create_tmpdir('config')

            pkg_repo_config, read_secret, write_secret = cls.setup_pkg_repo()
            pkg_repo_config_file = str(config_folder.join('config.toml'))
            admin_pkg_repo_secret_file = str(config_folder.join('admin_secret.toml'))

            BackendInstanceManager.dump_pkg_repo_configs(pkg_repo_config_file, [pkg_repo_config])
            BackendInstanceManager.dump_pkg_repo_secrets(admin_pkg_repo_secret_file, [write_secret])

            root_folder = str(create_tmpdir('root'))

            wstat = build_workflow_stat(
                    pkg_repo_config_file=pkg_repo_config_file,
                    admin_pkg_repo_secret_file=admin_pkg_repo_secret_file,
                    root_folder=root_folder,
                    auth_read_expires=0,
                    auth_write_expires=0,
            )

            return RepoInfoForTest(
                    name=pkg_repo_config.name.lower(),
                    pkg_repo_config_file=pkg_repo_config_file,
                    admin_pkg_repo_secret_file=admin_pkg_repo_secret_file,
                    root_folder=root_folder,
                    read_secret=read_secret,
                    write_secret=write_secret,
                    wstat=wstat,
            )

        @inject_to_caller
        @pytest.fixture(scope='session')
        def session_repo(tmpdir_factory):  # pylint: disable=unused-variable
            yield _create_repo_for_test(tmpdir_factory.mktemp)

        @inject_to_caller
        @pytest.fixture(scope='function')
        def function_repo(tmpdir):  # pylint: disable=unused-variable
            yield _create_repo_for_test(tmpdir.mkdir)

        @inject_to_caller
        @pytest.fixture(scope='function')
        def update_repo_index():  # pylint: disable=unused-variable
            yield cls.update_repo_index

        inject_to_caller(test_upload_with_write_secret)
        inject_to_caller(test_upload_with_read_secret)

        initialize_task_worker()



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
import functools
import hashlib
import importlib
import os
import pkgutil
import traceback
from typing import Callable, Dict, List, Tuple, Iterable, TypeVar, Type, Optional
import inspect

# TODO: follows https://github.com/PyCQA/pylint/issues/1524
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from private_pypi.utils import (
        read_toml,
        write_toml,
        normalize_distribution_name,
        update_hash_algo_with_file,
)


####################
# Static I/O types #
####################
class LocalPaths(BaseModel):
    index: str
    log: str
    lock: str
    job: str
    cache: str

    def makedirs(self):
        os.makedirs(self.index, exist_ok=True)
        os.makedirs(self.log, exist_ok=True)
        os.makedirs(self.lock, exist_ok=True)
        os.makedirs(self.job, exist_ok=True)
        os.makedirs(self.cache, exist_ok=True)


class UploadPackageStatus(Enum):
    SUCCEEDED = auto()
    FAILED = auto()


class UploadPackageResult(BaseModel):
    status: UploadPackageStatus
    message: str = ''


@dataclass
class UploadPackageContext:
    filename: str
    path: str
    meta: Optional[Dict[str, str]] = None
    failed: bool = False
    message: str = ''

    def __post_init__(self):
        # Fill distribution name.
        if not self.meta.get('distrib'):
            name = self.meta.get('name')
            if name:
                self.meta['distrib'] = normalize_distribution_name(name)
            else:
                self.failed = True
                self.message = 'Cannot generate the distribution name.'
        assert self.meta_distrib

        # SHA256 checksum, also suggested by PEP-503.
        if not self.meta.get('sha256'):
            sha256_algo = hashlib.sha256()
            update_hash_algo_with_file(self.path, sha256_algo)
            self.meta['sha256'] = sha256_algo.hexdigest()
        assert self.meta_sha256

    @property
    def meta_distrib(self) -> str:
        return self.meta['distrib']

    @property
    def meta_sha256(self) -> str:
        return self.meta['sha256']


class UploadIndexStatus(Enum):
    SUCCEEDED = auto()
    FAILED = auto()


class UploadIndexResult(BaseModel):
    status: UploadIndexStatus
    message: str = ''


class DownloadIndexStatus(Enum):
    SUCCEEDED = auto()
    FAILED = auto()


class DownloadIndexResult(BaseModel):
    status: DownloadIndexStatus
    message: str = ''


####################################
# Interfaces of package repository #
####################################
class PkgRepoConfig(BaseModel):
    type: str
    name: str
    max_file_bytes: int = 1024**3
    sync_index_interval: int = 60


class PkgRepoSecret(BaseModel):
    type: str
    name: str
    raw: str

    @abstractmethod
    def secret_hash(self) -> str:
        pass


class PkgRef(BaseModel):
    type: str
    distrib: str
    package: str
    ext: str
    sha256: str
    meta: Dict[str, str]

    @abstractmethod
    def auth_url(self, config: PkgRepoConfig, secret: PkgRepoSecret) -> str:
        pass


class PkgRepo(BaseModel):
    type: str
    config: PkgRepoConfig
    secret: PkgRepoSecret
    local_paths: LocalPaths

    @abstractmethod
    def record_error(self, error_message: str) -> None:
        pass

    @abstractmethod
    def ready(self) -> Tuple[bool, str]:
        pass

    @abstractmethod
    def auth_read(self) -> bool:
        pass

    @abstractmethod
    def auth_write(self) -> bool:
        pass

    @abstractmethod
    def upload_package(self, filename: str, meta: Dict[str, str], path: str) -> UploadPackageResult:
        pass

    @abstractmethod
    def collect_all_published_packages(self) -> List[PkgRef]:
        pass

    @abstractmethod
    def local_index_is_up_to_date(self, path: str) -> bool:
        pass

    @abstractmethod
    def upload_index(self, path: str) -> UploadIndexResult:
        pass

    @abstractmethod
    def download_index(self, path: str) -> DownloadIndexResult:
        pass


class BackendRegistration:
    type: str = ''
    pkg_repo_config_cls: Type[PkgRepoConfig] = PkgRepoConfig
    pkg_repo_secret_cls: Type[PkgRepoSecret] = PkgRepoSecret
    pkg_repo_cls: Type[PkgRepo] = PkgRepo
    pkg_ref_cls: Type[PkgRef] = PkgRef
    cli_name_to_func: Dict[str, Callable[[], int]] = {}


######################
# Backend reflection #
######################
class BackendInstanceManager:

    def __init__(self) -> None:
        self._type_to_registration = {}

        # Namespace package root.
        root_module = importlib.import_module('.', 'private_pypi_backends')
        # Find all submodules.
        for module_info in pkgutil.iter_modules(
                root_module.__path__,  # type: ignore
                root_module.__name__ + '.',
        ):
            # Load module.
            module = importlib.import_module(module_info.name)

            # Find the registration class.
            registration = None
            for obj in module.__dict__.values():
                if inspect.isclass(obj) \
                        and issubclass(obj, BackendRegistration) and obj is not BackendRegistration:
                    registration = obj

            if registration is None:
                continue

            # Type validation.
            assert registration.type

            assert issubclass(registration.pkg_repo_config_cls, PkgRepoConfig) \
                    and registration.pkg_repo_config_cls is not PkgRepoConfig

            assert issubclass(registration.pkg_repo_secret_cls, PkgRepoSecret) \
                    and registration.pkg_repo_secret_cls is not PkgRepoSecret

            assert issubclass(registration.pkg_repo_cls, PkgRepo) \
                    and registration.pkg_repo_cls is not PkgRepo

            assert issubclass(registration.pkg_ref_cls, PkgRef) \
                    and registration.pkg_ref_cls is not PkgRef

            self._type_to_registration[registration.type] = registration

    def _registration(self, **kwargs) -> Type[BackendRegistration]:
        assert 'type' in kwargs
        assert kwargs['type'] in self._type_to_registration
        return self._type_to_registration[kwargs['type']]

    def create_pkg_repo_config(self, **kwargs) -> PkgRepoConfig:
        return self._registration(**kwargs).pkg_repo_config_cls(**kwargs)

    def create_pkg_repo_secret(self, **kwargs) -> PkgRepoSecret:
        return self._registration(**kwargs).pkg_repo_secret_cls(**kwargs)

    def create_pkg_repo(self, **kwargs) -> PkgRepo:
        return self._registration(**kwargs).pkg_repo_cls(**kwargs)

    def create_pkg_ref(self, **kwargs) -> PkgRef:
        return self._registration(**kwargs).pkg_ref_cls(**kwargs)

    def load_pkg_repo_configs(self, path: str) -> Dict[str, PkgRepoConfig]:
        name_to_pkg_repo_config: Dict[str, PkgRepoConfig] = {}

        for name, struct in read_toml(path).items():
            if not isinstance(struct, dict):
                raise ValueError(f'Invalid pkg_repo_config, name={name}, struct={struct}')

            name = name.lower()
            config = self.create_pkg_repo_config(name=name, **struct)
            name_to_pkg_repo_config[name] = config

        return name_to_pkg_repo_config

    @staticmethod
    def dump_pkg_repo_configs(path: str, pkg_repo_configs: Iterable[PkgRepoConfig]) -> None:
        dump = {}

        for pkg_repo_config in pkg_repo_configs:
            struct = pkg_repo_config.dict()
            name = struct.pop('name')
            name = name.lower()
            dump[name] = struct

        write_toml(path, dump)

    def load_pkg_repo_secrets(self, path: str) -> Dict[str, PkgRepoSecret]:
        name_to_pkg_repo_secret: Dict[str, PkgRepoSecret] = {}

        for name, struct in read_toml(path).items():
            if not isinstance(struct, dict):
                raise ValueError(f'Invalid pkg_repo_secret, name={name}, struct={struct}')

            secret = self.create_pkg_repo_secret(name=name, **struct)
            name_to_pkg_repo_secret[name] = secret

        return name_to_pkg_repo_secret

    @staticmethod
    def dump_pkg_repo_secrets(path: str, pkg_repo_secrets: Iterable[PkgRepoSecret]) -> None:
        dump = {}

        for pkg_repo_config in pkg_repo_secrets:
            struct = pkg_repo_config.dict()
            name = struct.pop('name')
            name = name.lower()
            dump[name] = struct

        write_toml(path, dump)

    def load_pkg_refs(self, path: str) -> List[PkgRef]:
        return [self.create_pkg_ref(**struct) for struct in read_toml(path)['pkgs']]

    @staticmethod
    def dump_pkg_refs(path: str, pkg_refs: Iterable[PkgRef]) -> None:
        write_toml(path, {'pkgs': [pkg_ref.dict() for pkg_ref in pkg_refs]})


##########
# Helper #
##########
def basic_model_get_default(basic_model_cls: BaseModel, key: str):
    assert key in basic_model_cls.__fields__
    return basic_model_cls.__fields__[key].default


_METHOD = TypeVar('_METHOD')


def record_error_if_raises(method: _METHOD) -> _METHOD:

    @functools.wraps(method)
    def decorated(self, *args, **kwargs):
        try:
            ret = method(self, *args, **kwargs)
            return ret
        except:
            self.record_error(traceback.format_exc())
            raise

    return decorated


class PkgRepoIndex:

    def __init__(self, pkg_refs) -> None:
        self._distrib_to_pkg_refs: Dict[str, List[PkgRef]] = {}
        self._package_to_pkg_ref: Dict[str, PkgRef] = {}

        for pkg_ref in pkg_refs:
            self.add_pkg_ref(pkg_ref)

    def add_pkg_ref(self, pkg_ref: PkgRef) -> None:
        if pkg_ref.package in self._package_to_pkg_ref:
            raise KeyError(f'package={pkg_ref.package} duplicated.')

        if pkg_ref.distrib not in self._distrib_to_pkg_refs:
            self._distrib_to_pkg_refs[pkg_ref.distrib] = []

        self._distrib_to_pkg_refs[pkg_ref.distrib].append(pkg_ref)
        self._package_to_pkg_ref[pkg_ref.package] = pkg_ref

    @property
    def all_distributions(self) -> Iterable[str]:
        return self._distrib_to_pkg_refs.keys()

    def get_pkg_refs(self, query_distrib: str) -> Optional[List[PkgRef]]:
        distrib = normalize_distribution_name(query_distrib)
        return self._distrib_to_pkg_refs.get(distrib)

    def get_single_pkg_ref(self, query_distrib: str, query_package: str) -> Optional[PkgRef]:
        pkg_ref = self._package_to_pkg_ref.get(query_package)
        if pkg_ref is None or normalize_distribution_name(query_distrib) != pkg_ref.distrib:
            return None
        return pkg_ref
